# -*- Mode: Python; tab-width: 4; indent-tabs-mode: nil; -*-

from socketless.service import Protocol, Method

class InternalNodeServiceProtocol(Protocol):
    handshake = ('Tako Internal Node API Service', 'Tako Internal Node API Client')
    methods = dict(
        get  = Method('g', [('key', str)], [('timestamp', long), ('value', str)]), # key -> timestamp, value
        set  = Method('s', [('key', str), ('timestamp', long), ('value', str)], [('timestamp', long)]), # key, timestamp, value -> timestamp
        stat = Method('t', [('key', str)], [('timestamp', long)]), # key -> timestamp
    )

class PublicNodeServiceProtocol(Protocol):
    handshake = ('Tako Public Node API Service', 'Tako Public Node API Client')
    methods = dict(
        get  = Method('g', [('key', str)], [('timestamp', long), ('value', str)]), # key -> timestamp, value
        set  = Method('s', [('key', str), ('timestamp', long), ('value', str)], [('timestamp', long)]), # key, timestamp, value -> timestamp
        stat = Method('t', [('key', str)], [('timestamp', long)]), # key -> timestamp
    )




# -*- Mode: Python; tab-width: 4; indent-tabs-mode: nil; -*-

import hashlib
import bisect

class ConsistentHash(object):
    def __init__(self, buckets=[], points_per_bucket=120, buckets_per_key=3, points=None):
        super(ConsistentHash, self).__init__()
        self.buckets = list(buckets)
        self.points_per_bucket = points_per_bucket
        self.buckets_per_key = buckets_per_key
        self.points = []
        self.__neighbour_buckets_by_bucket_hash = {}
        self.__neighbour_buckets_by_point_index = []
        self.__bisect = bisect.bisect
        if points:
            self.points = points
        else:
            self.points = sorted([(self.__generate_point("%s-%s" % (hash(bucket), i)), bucket) for bucket in self.buckets for i in xrange(self.points_per_bucket)])
        self.__point_index = [point[0] for point in self.points]
        self.__point_count = len(self.points)
        for bucket in self.buckets:
            self.__neighbour_buckets_by_bucket_hash[hash(bucket)] = frozenset(self.__neighbour_buckets_for_bucket(bucket))
        for index, (coordinate, bucket) in enumerate(self.points):
            self.__neighbour_buckets_by_point_index.append(frozenset(self.__buckets_for_point_index(index)))

    def __generate_point(self, key):
        key_hash = hashlib.md5()
        key_hash.update(key)
        return int(key_hash.hexdigest()[:8], 16)

    def __buckets_for_point_index(self, index):
        buckets = set()
        buckets.add(self.points[index][1])
        n = 1
        while n < self.buckets_per_key and n < len(self.buckets):
            index += 1
            point, bucket = self.points[index % self.__point_count]
            buckets.add(bucket)
            n = len(buckets)
        return buckets

    def __neighbour_buckets_for_bucket(self, bucket):
        neighbour_buckets = set()
        bucket_points = [(i, p, b) for i, (p, b) in enumerate(self.points) if b == bucket]
        for i, p, b in bucket_points:
            point_neighbour_buckets = set()
            n = 0
            while n < self.buckets_per_key and n < len(self.buckets):
                i -= 1
                neighbour_point, neighbour_bucket = self.points[i % self.__point_count]
                point_neighbour_buckets.add(neighbour_bucket)
                n = len(point_neighbour_buckets)
            neighbour_buckets.update(point_neighbour_buckets)
        return neighbour_buckets

    def find_buckets(self, key):
        index = self.__bisect(self.__point_index, self.__generate_point(key))
        return self.__neighbour_buckets_by_point_index[index % self.__point_count]

    def find_neighbour_buckets(self, bucket):
        return self.__neighbour_buckets_by_bucket_hash[hash(bucket)]

    def __repr__(self):
        return 'ConsistentHash(buckets_per_key=%d, points_per_pucket=%d)' % (self.buckets_per_key, self.points_per_bucket)

    def __str__(self):
        return repr(self) + ''.join(['\n\t%08x: %s' % (point, str(bucket)) for point, bucket in self.points])



# -*- Mode: Python; tab-width: 4; indent-tabs-mode: nil; -*-

from datetime import timedelta
import logging
import os
import time
import urllib

from syncless import coio
from socketless import service

import paths
paths.setup()

from configurationcontroller import ConfigurationController
from protocols import InternalNodeServiceProtocol, PublicNodeServiceProtocol
from store import Store

class NoConfigurationException(BaseException):
    pass

class NodeServer(object):
    def __init__(self, node_id, store_file=None, explicit_configuration=None, coordinator_addresses=[], var_directory='var',
                 configuration_update_interval=300):
        super(NodeServer, self).__init__()
        if __debug__: logging.debug('node_id = %s, store_file = %s, explicit_configuration = %s, coordinators = %s, var_directory = %s',
                  node_id, store_file, explicit_configuration, coordinator_addresses, var_directory)
        self.id = node_id
        self.node = None
        var_directory = paths.path(var_directory)
        store_file = store_file or os.path.join(var_directory, 'data', '%s.tcb' % self.id)
        self.__store = Store(store_file)
        self.__store.open()
        self.__node_clients = {}
        self.__internal_cluster_client = service.MulticastClient(InternalNodeServiceProtocol())
        configuration_directory = os.path.join(var_directory, 'etc')
        self.__background_repair_enabled = False
        self.__background_repair_cycle_length_seconds = None
        self.__read_repair_enabled = False
        self.__configuration = None
        self.__configuration_controller = ConfigurationController('nodeserver-%s' % self.id,
                                                                  coordinator_addresses, explicit_configuration,
                                                                  configuration_directory, self.__update_configuration,
                                                                  configuration_update_interval)


    def __initialize_node_client_pool(self):
        nodes = self.__configuration.find_neighbour_nodes_for_node(self.node) if self.node else {}

        recycled_node_clients = {}
        for node_id, client in self.__node_clients.iteritems():
            if node_id in nodes:
                recycled_node_clients[node_id] = client
            else:
                client.disconnect()

        new_node_clients = {}
        for node_id, node in nodes.iteritems():
            if node_id not in recycled_node_clients:
                new_node_clients[node_id] = service.Client((node.address, node.port), InternalNodeServiceProtocol, tag=node_id)

        self.__node_clients = dict(recycled_node_clients, **new_node_clients)

        # Blocking, do this after setting self.__node_clients
        for client in new_node_clients.itervalues():
            client.connect()

    def __update_configuration(self, new_configuration):
        if __debug__: logging.debug('New configuration: %s', new_configuration)
        self.__configuration = new_configuration
        deployment = None
        if self.id in new_configuration.active_deployment.nodes:
            deployment = new_configuration.active_deployment
        if new_configuration.target_deployment and self.id in new_configuration.target_deployment.nodes:
            deployment = new_configuration.target_deployment
        self.node = deployment.nodes.get(self.id, None) if deployment else None
        self.__read_repair_enabled = deployment.read_repair_enabled if deployment else False
        self.__background_repair_enabled = deployment.background_repair_enabled if deployment else False
        self.__background_repair_interval_seconds = deployment.background_repair_interval_seconds if deployment else 0
        self.__initialize_node_client_pool()
        # TODO: restart servers if addresses changed

    def __quote(self, key):
        return urllib.quote_plus(key, safe='/&')

    def __unquote(self, path):
        return urllib.unquote_plus(path)

    def __fetch_value(self, key, node_id):
        if __debug__: logging.debug('key: %s, node_id: %s', key, node_id)
        return self.__clients_for_nodes((node_id,))[0].get(key) or (None, None)

    def __fetch_timestamps(self, key, node_ids):
        if __debug__: logging.debug('key: %s', key)
        node_ids = dict(node_ids)
        node_ids.pop(self.node.id, None)
        if not node_ids:
            return []
        clients = self.__clients_for_nodes(node_ids)
        return self.__internal_cluster_client.stat(clients, key)

    def __clients_for_nodes(self, node_ids):
        return [self.__node_clients[node_id] for node_id in node_ids]

    def __propagate(self, key, timestamp, value, target_nodes):
        if __debug__: logging.debug('key: %s, target_nodes: %s', key, target_nodes)
        collector = self.__internal_cluster_client.set_collector(self.__clients_for_nodes(target_nodes), 1)
        self.__internal_cluster_client.set_async(collector, key, timestamp, value)

    def __read_repair(self, key, timestamp, value, node_ids):
        if __debug__: logging.debug('key: %s, timestamp: %s', key, timestamp)
        remote_timestamps = self.__fetch_timestamps(key, node_ids)
        if __debug__: logging.debug('remote: %s', [(client.tag, repr(remote_timestamp)) for client, remote_timestamp in remote_timestamps])
        newer = [(client, remote_timestamp) for client, remote_timestamp in remote_timestamps
                 if remote_timestamp and remote_timestamp > timestamp]

        if __debug__: logging.debug('newer: %s', [(client.tag, repr(remote_timestamp)) for client, remote_timestamp in newer])
        if newer:
            latest_client, latest_timestamp = newer[-1]
            latest_timestamp, latest_value = self.__fetch_value(key, latest_client.tag)
            if __debug__: logging.debug('latest_timestamp: %s', latest_timestamp)
            if latest_timestamp and latest_value:
                value = latest_value
                timestamp = latest_timestamp

        older = [(client, remote_timestamp) for client, remote_timestamp in remote_timestamps
                 if remote_timestamp != None and remote_timestamp < timestamp]
        if __debug__: logging.debug('older: %s', [(client.tag, repr(remote_timestamp)) for client, remote_timestamp in older])
        if older:
            older_node_ids = [client.tag for (client, remote_timestamp) in older]
            self.__propagate(key, timestamp, value, older_node_ids)

        return timestamp, value

    def __internal_get(self, key):
        if __debug__: logging.debug('key: %s', key)
        timestamp, value = self.__store.get(key)
        return timestamp or 0, value

    def __internal_set(self, key, timestamp, value):
        if __debug__: logging.debug('key: %s', key)
        self.__store.set(key, timestamp, value)
        return timestamp

    def __internal_stat(self, key):
        if __debug__: logging.debug('key: %s', key)
        timestamp, value = self.__store.get(key)
        if __debug__: logging.debug('timestamp: %s', timestamp)
        return timestamp or 0

    def __public_get(self, key):
        if __debug__: logging.debug("key: %s", key)
        timestamp, value = None, None
        node_ids = self.__configuration.find_nodes_for_key(key)
        if self.node.id in node_ids:
            timestamp, value = self.__store.get(key)
        if self.__read_repair_enabled:
            new_timestamp, new_value = self.__read_repair(key, timestamp, value, node_ids)
            if new_timestamp > timestamp and self.node.id in node_ids:
                timestamp, value = new_timestamp, new_value
                self.__store.set(key, timestamp, value)
        return timestamp or 0, value


    def __public_set(self, key, timestamp, value):
        if __debug__: logging.debug("key: %s", key)
        node_ids = self.__configuration.find_nodes_for_key(key)
        if self.__read_repair_enabled:
            timestamp, value = self.__read_repair(key, timestamp, value, node_ids)
        if self.node.id in node_ids:
            self.__store.set(key, timestamp, value)
        return timestamp

    def __public_stat(self, key):
        return self.__public_get(key)

    def __repair_key(self, key):
        node_ids = self.__configuration.find_nodes_for_key(key)
        if self.node.id not in node_ids:
            self.__store.remove(key)
        else:
            timestamp, value = self.__store.get(key)
            if timestamp:
                new_timestamp, new_value = self.__read_repair(key, timestamp, value, node_ids)
                if new_timestamp > timestamp:
                    self.__store.set(key, new_timestamp, new_value)

    def __repair_store(self):
        if not self.node:
            return
        if not self.__background_repair_enabled:
            return

        total_count = self.__store.count()
        scan_count = 0
        start_time = time.time()
        last_time = start_time
        elapsed_time = 0

        logging.info('Starting store repair. Total %d keys.', total_count)

        # TODO: Refactor away this cursor muck
        cursor = self.__store.cursor()
        while True:
            key = None
            try:
                cursor.next()
                key = cursor.key()
            except StopIteration:
                break
            except KeyError, e:
                logging.warning('KeyError: %s', e)
            except Exception, e:
                logging.warning('TC Exception: %s', e)
                # Not supposed to get this... pytc is broken.
                pass
            if not key:
                continue
            self.__repair_key(key)
            scan_count += 1
            now = time.time()
            if now - last_time > 5.0:
                last_time = now
                elapsed_time = now - start_time
                if __debug__: logging.debug('Store repair in progress. Scanned %d keys. Elapsed time: %s', scan_count, timedelta(seconds=elapsed_time))
            total_count = self.__store.count()
            if total_count == 0:
                break
            if not self.node:
                break
            if not self.__background_repair_enabled:
                break
            coio.sleep(self.__background_repair_interval_seconds / total_count)

        logging.info('Finished store repair. Scanned %d keys. Elapsed time: %s', scan_count, timedelta(seconds=elapsed_time))

    def __repair_task_loop(self):
        while True:
            start_time = time.time()
            try:
                self.__repair_store()
            except Exception, e:
                logging.exception(e)
                pass
            elapsed = time.time() - start_time
            spare_seconds = self.__background_repair_interval_seconds - elapsed
            if spare_seconds > 0:
                logging.info('Repair task sleeping %s', timedelta(seconds=spare_seconds))
                coio.sleep(spare_seconds)

    def serve(self):
        self.__configuration_controller.start()
        while not self.__configuration:
            if __debug__: logging.debug('Waiting for configuration.')
            coio.sleep(1)

        internal_service = service.Service(
            InternalNodeServiceProtocol(),
            get=self.__internal_get,
            set=self.__internal_set,
            stat=self.__internal_stat,
        )

        public_service = service.Service(
            PublicNodeServiceProtocol(),
            get=self.__public_get,
            set=self.__public_set,
            stat=self.__public_stat,
        )

        self.__repair_task = coio.stackless.tasklet(self.__repair_task_loop)()

        logging.info('Internal API: %s:%s' % (self.node.address, self.node.port))
        self.__server = service.Server(listener=('', self.node.port),
                                                services=(internal_service, public_service))
        self.__server.serve()

    def stop(self):
        self.__server.stop()
        self.__server = None



# -*- Mode: Python; tab-width: 4; indent-tabs-mode: nil; -*-

from __future__ import with_statement

import yaml
import simplejson as json
import logging
import os

import paths
paths.setup()

from utils import timestamper
from models import Coordinator, Deployment

def try_load_representation(representation, timestamp=timestamper.now()):
    if __debug__: logging.debug('timestamp = %s, representation = %s', timestamp, representation)
    try:
        configuration = Configuration(representation, timestamp)
        return configuration
    except ValidationError, e:
        logging.error('Configuration is not valid: %s', e)
        return None

def try_load_json(s, timestamp=timestamper.now()):
    try:
        representation = json.loads(s)
    except Exception, e:
        logging.error('Failed to parse JSON configuration: %s', e)
        return None
    return try_load_representation(representation, timestamp)

def try_load_file(filepath, timestamp=None):
    try:
        if not timestamp:
            timestamp = timestamper.from_seconds(os.path.getmtime(filepath))
        with open(filepath) as f:
            representation = yaml.load(f)
    except OSError, e:
        logging.error('Failed reading configuration file: %s', e)
        return None
    except IOError, e:
        logging.error('Failed reading configuration file: %s', e)
        return None
    except Exception, e:
        logging.error('Failed reading configuration file: %s', e)
        return None
    return try_load_representation(representation, timestamp)

def try_dump_file(filepath, configuration):
    if __debug__: logging.debug(filepath)
    try:
        dirpath = os.path.dirname(filepath)
        if not os.path.exists(dirpath):
            os.makedirs(dirpath)
        with open(filepath, 'w+') as f:
            yaml.dump(configuration.representation(), f)
        return True
    except IOError, e:
        logging.error('Failed to write configuration to file: %s', e)
        return False
    except OSError, e:
        logging.error('Failed to write configuration to file: %s', e)
        return False

def validate_representation(representation):
    # TODO: Complete validation
    # TODO: Validate that node id's are not recycled/nodes are not changed
    if not 'active_deployment' in representation:
        raise ValidationError('Missing active_deployment.')
    if not 'deployments' in representation:
        raise ValidationError('Missing deployments.')
    if not len(representation['deployments']) > 0:
        raise ValidationError('deployments list is empty.')
    active_deployment_id = representation['active_deployment']
    if not active_deployment_id in representation['deployments']:
        raise ValidationError('Active deployment %s not present in deployment list.' % active_deployment_id)
    if 'target_deployment' in representation:
        target_deployment_id = representation['target_deployment']
        if not target_deployment_id in representation['deployments']:
            raise ValidationError('Target deployment %s not present in deployment list.' % target_deployment_id)
    for deployment_id, deployment in representation['deployments'].iteritems():
        if not 'buckets' in deployment:
            raise ValidationError('Missing buckets in deployment %s' % deployment_id)
        for bucket_id, bucket in deployment['buckets'].iteritems():
            if not len(bucket) > 0:
                raise ValidationError('No nodes in bucket %s in deployment %s' % (bucket_id, deployment_id))
            for node_id, node in bucket.iteritems():
                if not len(node) == 2:
                    raise ValidationError('Malformed node %s in bucket %s in deployment %s' % (node_id, bucket_id, deployment_id))
                address, port = node
                if not type(address) == str:
                    raise ValidationError('Malformed node %s in bucket %s in deployment %s' % (node_id, bucket_id, deployment_id))
                if not type(port) == int:
                    raise ValidationError('Malformed node %s in bucket %s in deployment %s' % (node_id, bucket_id, deployment_id))
    if 'master_coordinator' in representation:
        master_coordinator_id = representation['master_coordinator']
        if not master_coordinator_id in representation['coordinators']:
            raise ValidationError('Master coordinator %s not present in coordinators list.' % master_coordinator_id)
    for coordinator_id, coordinator in representation.get('coordinators', {}).iteritems():
        if not len(coordinator) == 2: raise ValidationError()
        address, port = coordinator
        if not type(address) == str: raise ValidationError('type(address) == %s != str' % type(address))
        if not type(port) == int: raise ValidationError('type(port) == %s != int' % type(port))

class ValidationError(Exception):
    def __init__(self, description=None):
        super(ValidationError, self).__init__()
        self.description = description

    def __str__(self):
        return 'ValidationError(%s)' % (self.description or '')

    def __repr__(self):
        return str(self)

class Configuration(object):
    def __init__(self, representation=None, timestamp=timestamper.now()):
        super(Configuration, self).__init__()
        self.timestamp = timestamp
        if representation:
            validate_representation(representation)
            self.original_representation = representation
            self.deployments = dict((name, Deployment(name, deployment_representation)) for name, deployment_representation in representation.get('deployments', {}).iteritems())
            self.active_deployment_name = representation['active_deployment']
            self.active_deployment = self.deployments[self.active_deployment_name]
            self.target_deployment_name = representation.get('target_deployment', None)
            self.target_deployment = self.deployments[self.target_deployment_name] if self.target_deployment_name else None
            self.coordinators = dict((coordinator_id, Coordinator(coordinator_id, address, port)) for coordinator_id, (address, port) in representation.get('coordinators', {}).iteritems())
            self.master_coordinator_id = representation.get('master_coordinator', None)
            self.master_coordinator = self.coordinators.get(self.master_coordinator_id, None)

    def __repr__(self):
        return str(self)

    def __str__(self):
        return "Configuration(%s)" % self.representation()

    def representation(self):
        spec = {
                'active_deployment': self.active_deployment_name,
                'deployments': dict((deployment.name, deployment.representation()) for deployment in self.deployments.itervalues()),
        }
        if self.coordinators:
            spec['coordinators'] = dict((coordinator.id, [coordinator.address, coordinator.port]) for coordinator in self.coordinators.itervalues())
        if self.master_coordinator:
            spec['master_coordinator'] = self.master_coordinator_id
        if self.target_deployment_name:
            spec['target_deployment'] = self.target_deployment_name
        return spec

    def find_nodes_for_key(self, key):
        nodes = dict([(node.id, node) for bucket in self.active_deployment.buckets_for_key(key) for node in bucket])
        if self.target_deployment:
            nodes.update(dict([(node.id, node) for bucket in self.target_deployment.buckets_for_key(key) for node in bucket]))
        return nodes

    def find_neighbour_nodes_for_node(self, local_node):
        neighbour_buckets = []
        if local_node.id in self.active_deployment.nodes:
            deployment_node_bucket = self.active_deployment.buckets.get(local_node.bucket_id, None)
            neighbour_buckets.extend(self.active_deployment.consistent_hash.find_neighbour_buckets(deployment_node_bucket))
        if self.target_deployment and local_node.id in self.target_deployment.nodes:
            target_node_bucket = self.target_deployment.buckets.get(local_node.bucket_id, None)
            neighbour_buckets.extend(self.target_deployment.consistent_hash.find_neighbour_buckets(target_node_bucket))
        neighbour_nodes = dict((node.id, node) for bucket in neighbour_buckets for node in bucket)
        neighbour_nodes.pop(local_node.id, None)
        return neighbour_nodes

    def all_nodes(self):
        nodes = dict((node.id, node) for bucket in self.active_deployment.buckets.itervalues() for node in bucket)
        if self.target_deployment:
            nodes.update([(node.id, node) for bucket in self.target_deployment.buckets.itervalues() for node in bucket])
        return nodes



# -*- Mode: Python; tab-width: 4; indent-tabs-mode: nil; -*-
import logging

import paths
paths.setup()

from socketless import service
from syncless import coio

from configurationcontroller import ConfigurationController
from protocols import PublicNodeServiceProtocol

class NotAvailableException(BaseException):
    def __init__(self, key):
        super(NotAvailableException, self).__init__('Value not currently available for key "%s"' % repr(key))
        self.key = key

class Client(object):
    def __init__(self, name, coordinator_addresses=[], explicit_configuration=None, configuration_cache_directory=None,
                 configuration_update_interval=5*30, max_retries=30, retry_interval=1.0):
        super(Client, self).__init__()
        self.max_retries = max_retries
        self.retry_interval = retry_interval
        self.__node_clients = {}
        self.__configuration = None
        self.__configuration_controller = ConfigurationController(name, coordinator_addresses, explicit_configuration,
                                                                  configuration_cache_directory, self.__update_configuration,
                                                                  configuration_update_interval)

    def __initialize_node_client_pool(self):
        nodes = self.__configuration.all_nodes()

        recycled_node_clients = {}
        for node_id, client in self.__node_clients.iteritems():
            if node_id in nodes:
                recycled_node_clients[node_id] = client
            else:
                client.disconnect()

        new_node_clients = {}
        for node_id, node in nodes.iteritems():
            if node_id not in recycled_node_clients:
                new_node_clients[node_id] = service.Client((node.address, node.port), PublicNodeServiceProtocol, tag=node_id)

        self.__node_clients = dict(recycled_node_clients, **new_node_clients)

        # Blocking, do this after setting self.__node_clients
        for client in new_node_clients.itervalues():
            client.connect()

    def __update_configuration(self, new_configuration):
        if __debug__: logging.debug(new_configuration)
        self.__configuration = new_configuration
        self.__initialize_node_client_pool()

    def __connected_clients_for_key(self, key):
        # TODO: Distribute node selection evenly?
        if not self.__node_clients or not self.__configuration:
            return []
        nodes = self.__configuration.find_nodes_for_key(key)
        clients = [self.__node_clients[node_id] for node_id in nodes.keys()]
        connected_clients = [client for client in clients if client.is_connected()]
        if __debug__: logging.debug('connected_clients = %s', connected_clients)
        return connected_clients

    def connect(self):
        self.__configuration_controller.start()

    def disconnect(self):
        self.__configuration_controller.stop()
        for node_id, client in self.__node_clients.iteritems():
            client.disconnect()
        self.__node_clients = None

    def connected_node_count(self):
        return len([node_client for node_client in self.__node_clients.itervalues() if node_client.is_connected()])

    def total_node_count(self):
        return len(self.__node_clients)

    def has_configuration(self):
        return self.__configuration_controller.configuration != None

    def is_connected(self, complete=False):
        if not complete:
            return self.has_configuration() and self.connected_node_count() > self.total_node_count() / 2.0
        else:
            return self.has_configuration() and self.connected_node_count() == self.total_node_count();

    def set(self, key, timestamp, value):
        if __debug__: logging.debug('key = %s, timestamp = %s', key, timestamp)
        for i in xrange(self.max_retries + 1):
            node_clients = self.__connected_clients_for_key(key)
            for node_client in node_clients:
                new_timestamp = node_client.set(key, timestamp, value)
                if new_timestamp is not None:
                    return new_timestamp
            coio.sleep(self.retry_interval)
        raise NotAvailableException(key)

    def get(self, key):
        if __debug__: logging.debug('key = %s', key)
        for i in xrange(self.max_retries + 1):
            node_clients = self.__connected_clients_for_key(key)
            for node_client in node_clients:
                result = node_client.get(key)
                if __debug__: logging.debug('result = %s', result)
                if result is not None:
                    timestamp, value = result
                    return timestamp or None, value
            coio.sleep(self.retry_interval)
        raise NotAvailableException(key)

    def stat(self, key):
        if __debug__: logging.debug('key = %s', key)
        for i in xrange(self.max_retries + 1):
            node_clients = self.__connected_clients_for_key(key)
            for node_client in node_clients:
                timestamp = node_client.stat(key)
                if timestamp is not None:
                    return timestamp
            coio.sleep(self.retry_interval)
        raise NotAvailableException(key)



# -*- Mode: Python; tab-width: 4; indent-tabs-mode: nil; -*-

import logging.handlers
import logging
import pprint

def whoami():
    import sys
    return sys._getframe(1).f_code.co_name

def callersname():
    import sys
    return sys._getframe(2).f_code.co_name

def pp():
    return pprint.PrettyPrinter(indent=4)

def configure_logging(appname, level=logging.INFO, filename=None):
    if level <= logging.DEBUG:
        format = "%(asctime)-15s %(levelname)s (%(process)d) %(filename)s:%(lineno)d %(funcName)s(): %(message)s"
    else:
        format="%(asctime)-15s %(levelname)s: %(message)s"
    logging.basicConfig(level=level, filename=filename, format=format)



# -*- Mode: Python; tab-width: 4; indent-tabs-mode: nil; -*-

import logging
import random

import pytc as tc
from syncless import coio

from utils import debug
from utils import timestamper

BUFFER_THRESHOLD = 4096

class Store(object):
    def __init__(self, filepath, auto_commit_interval=2):
        self.operation_counter = 0
        super(Store, self).__init__()
        if __debug__: logging.debug('filepath: %s', filepath)
        self.filepath = filepath
        self.db = tc.BDB()
        self.db.tune(0, 0, 1024**2*10, -1, -1, tc.BDBTLARGE)
        self.flusher = None
        self.auto_commit_interval = auto_commit_interval
        self.pack_timestamp = timestamper.pack
        self.unpack_timestamp = timestamper.unpack

    def open(self):
        self.db.open(self.filepath, tc.BDBOWRITER | tc.BDBOCREAT)
        if self.auto_commit_interval:
            self.begin()
            self.flusher = coio.stackless.tasklet(self.__flush)()

    def close(self):
        if self.auto_commit_interval:
            self.flusher.kill()
            self.flusher = None
            self.commit()
        self.db.close()

    def __flush(self):
        coio.sleep(random.random() * self.auto_commit_interval)
        while True:
            if self.operation_counter > 0:
                if __debug__: logging.debug('Committing %d operations', self.operation_counter)
                self.commit()
                self.operation_counter = 0
                # Close and reopen to free memory allocated by TC
                # Otherwise memory usage balloons until we run out of memory and get killed
                self.db.close()
                self.db.open(self.filepath, tc.BDBOWRITER | tc.BDBOCREAT)
                self.begin()
            coio.sleep(self.auto_commit_interval)

    def __unpack_timestamped_data(self, data):
        if len(data) > BUFFER_THRESHOLD:
            value = buffer(data, 8)
        else:
            value = data[8:]
        timestamp = self.unpack_timestamp(data[0:8])
        return timestamp, value

    def __jump(self, cur, start):
        keylen = len(start)
        cur.jump(start)
        key = cur.key()
        if keylen:
            while not key[:keylen] > start:
                cur.next()
                key = cur.key()
        return key

    def set(self, key, timestamp, value):
        self.operation_counter += 1
        self.db.put(key, self.pack_timestamp(timestamp))
        self.db.putcat(key, value)

    def get(self, key):
        self.operation_counter += 1
        try:
            data = self.db.get(key)
            return self.__unpack_timestamped_data(data)
        except Exception:
            pass
        return (None, None)

    def remove(self, key):
        try:
            self.db.out(key)
        except Exception:
            pass

    def cursor(self, start_key=None):
        cur = self.db.curnew()
        try:
            if start_key:
                self.__jump(cur, start_key)
            else:
                cur.first()
        except KeyError:
            pass
        return cur

    def abort(self):
        self.db.tranabort()

    def begin(self):
        self.db.tranbegin()

    def commit(self):
        self.db.trancommit()

    def count(self):
        # Fast, calls rnum
        return len(self.db)




# -*- Mode: Python; tab-width: 4; indent-tabs-mode: nil; -*-

from setuptools import setup

setup(
    name='tako',
    version='0.0.25',
    packages=['tako', 'tako.utils'],
    scripts=[
        'bin/tako-cluster',
        'bin/tako-coordinator',
        'bin/tako-node',
        'bin/tako-proxy'
    ],

    install_requires = [
        'socketless >= 0.4.0',
        'pytc >= 0.8',
        'argparse',
        'pyyaml',
        'simplejson',
        'urllib3'
        ],
    zip_safe=True,

    author = "Daniel Norberg",
    author_email = "daniel.norberg@gmail.com",
    url = "https://github.com/danielnorberg/tako/",
    description = "Tako: A distributed data store.",
        long_description =      """\
        Tako is a distributed key-value data store. It aims to provide high
        scalability and availability through a shared nothing architecture,
        consistent hash based data partitioning, read repair with time
        stamping and live migration. An included coordinator server can be
        used as a single point of configuration to distribute cluster
        configuration to tako nodes in a cluster.

        Tako includes a http proxy server that can be used to
        interface with a tako cluster using normal HTTP GET/POST.

        Within a cluster, tako nodes communicate using a binary
        protocol and the socketless and syncless libraries.

        Tako makes use of libev/libevent/kqueue/kpoll if present.
        Tokyo Cabinet is used for data storage.

        Project Home: https://github.com/danielnorberg/tako
        """,
    license="Apache License, Version 2.0",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Environment :: Console",
        "Environment :: No Input/Output (Daemon)",
        "Environment :: Other Environment",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: POSIX :: Linux",
        "Operating System :: Unix",
        "Programming Language :: Python :: 2.5",
        "Programming Language :: Python :: 2.6",
        "Programming Language :: Python :: 2.7",
        "Topic :: Internet",
        "Topic :: Software Development :: Libraries :: Application Frameworks",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
)



# -*- Mode: Python; tab-width: 4; indent-tabs-mode: nil; -*-

import logging
import socket

from syncless import coio
from syncless import wsgi

class BadRequest(object):
    def __init__(self, description=''):
        super(BadRequest, self).__init__()
        self.description = description

    def __str__(self):
        return repr(self)

    def __repr__(self):
        return "BadRequest('%s')" % self.description

class HttpServer(object):
    def __init__(self, listener, handlers, listen_queue_size=1024):
        super(HttpServer, self).__init__()
        self.address, self.port = listener
        self.handlers = handlers
        self.listen_queue_size = listen_queue_size
        self.server_socket = None

    def handle_request(self, env, start_response):
        method = env['REQUEST_METHOD']
        path = env['PATH_INFO']
        body = env['wsgi.input']
        for prefix, methods in self.handlers:
            if path.startswith(prefix):
                sub_path = path[len(prefix):]
                method_handler = methods.get(method, None)
                if method_handler:
                    try:
                        return method_handler(start_response, sub_path, body, env)
                    except BadRequest, e:
                        start_response('400 Bad Request', [e.description])
                        return ['']
                else:
                    start_response('405 Method Not Allowed', [('Allow', ','.join(methods.keys()))])
        start_response('404 Not Found', [])
        return ['']

    def serve(self):
        self.server_socket = coio.nbsocket(socket.AF_INET, socket.SOCK_STREAM)
        self.server_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.server_socket.bind((self.address, self.port))
        self.server_socket.listen(self.listen_queue_size)
        if __debug__: logging.debug('listening on %r' % (self.server_socket.getsockname(),))
        wsgi.WsgiListener(self.server_socket, self.handle_request)



# -*- Mode: Python; tab-width: 4; indent-tabs-mode: nil; -*-

import re
from datetime import timedelta

def parse_timedelta(s):
    if s is None:
        return None
    d = re.match(
        r'((?P<days>\d+)\s*(days|day|d)\s*,?)?\s*((?P<hours>\d+)\s*:\s*'
        r'(?P<minutes>\d+)\s*:\s*(?P<seconds>\d+))?\s*',
        str(s)).groupdict(0)
    return timedelta(**dict(((key, int(value)) for key, value in d.items())))
